import { HAP, Logging, Service } from 'homebridge'
import * as HomeKitTypes from 'hap-nodejs/dist/lib/gen'

export type Config = {
  reConnectThreshold: number // 90000
  reConnectFailureThreshold: number // 10000
  requestConnectDebounceThreshold: number // 40
  requestPropertyDebounceThreshold: number // 100
}

export type MIoTSpec = { name: string; siid: number, piid: number }
export type MIoTSpecs = MIoTSpec[]
export type MIoTSpecMapping = { [name: string]: MIoTSpec }
export type MIoTSpecGetQuery = MIoTSpec & { did: string }
export type MIoTSpecsGetQuery = MIoTSpecGetQuery[]
export type MIoTSpecSetQuery = MIoTSpec & { did: string; value: any }
export type MIoTSpecsSetQuery = MIoTSpecSetQuery[]
export type MIoTSpecsResponse<T> = MIoTSpec & { did: string; value: T }
export type MIoTSpecsResponseMapping = { [name: string]: MIoTSpecsResponse<any> }
export type MIoTSpecsResponseValueMapping = { [name: string]: any }
export type MIoTDeviceIdentify = { name: string; token: string; address: string; }

export type MIIOSpec = string
export type MIIOSpecMapping = { [name: string]: MIIOSpec }
export type MIIOSpecGetQuery = MIIOSpec[]
export type MIIOSpecResponse = (string | number)[]
export type MIIOSpecResponseValueMapping = { [name: string]: any }

export type DeviceConfig = { name: string; address: string; token: string; }
export type DeviceConfigs = DeviceConfig[]

export type PropertyKey = string
export type PropertyValue = string | number | boolean;

export class DeviceInstance {
  public id: string
  public did: string
  public timeout: number
  // MIoT
  public miioCall<T extends any> (action: 'get_properties', specQuery: MIoTSpecsGetQuery): MIoTSpecsResponse<T>[]
  public miioCall<T extends any> (action: 'set_properties', specQuery: MIoTSpecsSetQuery): MIoTSpecsResponse<T>[]
  // MIIO
  public miioCall (action: 'get_prop', specQuery: MIIOSpecGetQuery): MIIOSpecResponse
  public miioCall (action: string, specQuery: PropertyValue[]): any
}


export function initMIoT (props: { log: Logging; hap: HAP; config?: Config }): void


export type HomeKitTypes = ValueOf<typeof HomeKitTypes.Generated>
export type MIoTDeviceProps = {
  identify: MIoTDeviceIdentify
  characteristicsService: Service
}
export type MIoTRegisterConfig = {
  name?: string,
  get: { formatter: (valueMapping: MIoTSpecsResponseValueMapping) => any }
  set?: { formatter: (value: any, previousProperty: MIoTSpecsResponseValueMapping) => PropertyValue; property: PropertyKey }
}
export type MIIORegisterConfig = {
  name?: string,
  get: { formatter: (valueMapping: MIoTSpecsResponseValueMapping) => any }
  set?: { formatter: (value: any, previousProperty: MIIOSpecResponseValueMapping) => PropertyValue[]; property: PropertyKey }
}
export class MIoTDevice {
  // Constructor
  constructor (props: MIoTDeviceProps)
  // MIoT
  public addMIoTSpec (specs: MIoTSpecMapping): void
  public addMIoTCharacteristicListener(type: HomeKitTypes, config: MIoTRegisterConfig): void
  // MIIO
  public addMIIOSpec (specs: MIIOSpecMapping): void
  public addMIIOCharacteristicListener(type: HomeKitTypes, config: MIIORegisterConfig): void
}


export class Foundation {
  public log: Logging
  public hap: HAP
  public config: Config
}
export const SharedFoundation: InstanceType<typeof Foundation>


type ValueOf<T> = T[keyof T]
export type Resolver = (...args: any[]) => void
